import pytest
from django.core.exceptions import ImproperlyConfigured

from django_iseries.creation import DatabaseCreation
from tests.models import Object, ObjectReference


@pytest.mark.django_db
def test_sanity_check():
    assert True


@pytest.fixture
def connection():
    from django.db import connection
    new_connection = connection.copy()
    yield new_connection
    new_connection.close()


@pytest.fixture
def editor(connection):
    return connection.schema_editor()


@pytest.mark.parametrize('value,expected', [
    ('string', "'string'"),
    (42, '42'),
    (1.754, '1.754'),
    (False, '0'),
])
def test_quote_value(editor, value, expected):
    actual = editor.quote_value(value)
    assert expected == actual


# noinspection PyProtectedMember
def test_create_test_db(connection):
    creation = DatabaseCreation(connection)
    with pytest.raises(ImproperlyConfigured):
        # Db2 for iSeries does not support creating new databases
        creation._create_test_db(verbosity=0, autoclobber=False, keepdb=False)
    creation._create_test_db(verbosity=0, autoclobber=False, keepdb=True)


class DBConstraintTests:
    @pytest.mark.django_db
    def test_can_reference_existent(self):
        obj = Object.objects.create()
        ref = ObjectReference.objects.create(obj=obj)
        assert ref.obj == obj

        ref = ObjectReference.objects.get(obj=obj)
        assert ref.obj == obj

    @pytest.mark.django_db
    def test_can_reference_non_existent(self):
        assert not Object.objects.filter(id=12345).exists()
        ref = ObjectReference.objects.create(obj_id=12345)
        ref_new = ObjectReference.objects.get(obj_id=12345)
        assert ref == ref_new

        # noinspection PyTypeChecker
        with pytest.raises(Object.DoesNotExist):
            # noinspection PyStatementEffect
            ref.obj

    @pytest.mark.django_db
    def test_many_to_many(self):
        obj = Object.objects.create()
        obj.related_objects.create()
        assert Object.objects.count() == 2
        assert obj.related_objects.count() == 1

        intermediary_model = Object._meta.get_field("related_objects").remote_field.through
        intermediary_model.objects.create(from_object_id=obj.id, to_object_id=12345)
        assert obj.related_objects.count() == 1
        assert intermediary_model.objects.count() == 2



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                      |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi,             |
# | Hemlata Bhatt, Vyshakh A                                                 |
# +--------------------------------------------------------------------------+

import datetime
import uuid
from functools import lru_cache

from django.conf import settings
from django.db import utils
from django.db.backends.base.operations import BaseDatabaseOperations
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.timezone import is_aware, utc

from django_iseries import query

dbms_name = 'dbms_name'


class DatabaseOperations(BaseDatabaseOperations):
    def __init__(self, connection):
        super(DatabaseOperations, self).__init__(self)
        self.connection = connection

    compiler_module = "django_iseries.compiler"

    def cache_key_culling_sql(self):
        return '''select cache_key 
                    from (select cache_key, ( ROW_NUMBER() over() ) as rownum from %s order by cache_key)
                    where rownum = %%s + 1
        '''

    def check_aggregate_support(self, aggregate):
        # In DB2 data type of the result is the same as the data type of the argument values for AVG aggregation
        # But Django aspect in Float regardless of data types of argument value
        # http://publib.boulder.ibm.com/infocenter/db2luw/v9r7/index.jsp?topic=/com.ibm.db2.luw.apdv.cli.doc/doc/c0007645.html
        if aggregate.sql_function == 'AVG':
            aggregate.sql_template = '%(function)s(DOUBLE(%(field)s))'
        # In DB2 equivalent sql function of STDDEV_POP is STDDEV
        elif aggregate.sql_function == 'STDDEV_POP':
            aggregate.sql_function = 'STDDEV'
        # In DB2 equivalent sql function of VAR_SAMP is VARIENCE
        elif aggregate.sql_function == 'VAR_POP':
            aggregate.sql_function = 'VARIANCE'
        # DB2 doesn't have sample standard deviation function
        elif aggregate.sql_function == 'STDDEV_SAMP':
            raise NotImplementedError("sample standard deviation function not supported")
        # DB2 doesn't have sample variance function
        elif aggregate.sql_function == 'VAR_SAMP':
            raise NotImplementedError("sample variance function not supported")

    def get_db_converters(self, expression):
        """
        Get a list of functions needed to convert field data.
        """
        converters = super(DatabaseOperations, self).get_db_converters(expression)

        internal_type = expression.output_field.get_internal_type()
        if internal_type == 'UUIDField':
            converters.append(self.convert_uuidfield_value)
        return converters

    def combine_duration_expression(self, connector, sub_expressions):
        if 'NULL' in sub_expressions:
            # django expects that adding a null duration will return null
            # db2 does not allow untyped columns
            return 'CAST(NULL AS TIMESTAMP)'

        lhs, rhs = sub_expressions
        if connector not in '+-':
            raise utils.DatabaseError('Invalid connector for timedelta: %s.' % connector)
        if 'MICROSECONDS' not in lhs:
            lhs = f'CAST({lhs} as TIMESTAMP)'
        if 'MICROSECONDS' not in rhs:
            rhs = f'CAST({rhs} as TIMESTAMP)'
            lhs, rhs = rhs, lhs
        if connector == '-' and 'MICROSECONDS' in rhs:
            rhs = rhs.replace('+', '-')
        return f'{lhs} {connector} {rhs}'

    def combine_expression(self, connector, sub_expressions):
        lhs, rhs = sub_expressions
        if connector == '%%':
            # Db2 for iSeries requires explicit cast for parameters to MOD. INTEGER is the only cast that makes sense
            # (though this may cause failures for BIGINT types down the road)
            return f'MOD(CAST({lhs} AS INTEGER), CAST({rhs} AS INTEGER))'
        elif connector == '&':
            return f'BITAND({lhs}, {rhs})'
        elif connector == '|':
            return f'BITOR({lhs}, {rhs})'
        elif connector == '<<':
            return f'({lhs} * POWER(2, CAST({rhs} AS INTEGER)))'
        elif connector == '>>':
            return f'FLOOR({lhs} / POWER(2, CAST({rhs} AS INTEGER)))'
        elif connector == '^':
            # Db2 for iSeries requires explicit cast for parameters to POWER. FLOAT is chosen here as it will return
            # correct values, though it may sacrifice a small amount of speed
            return f'POWER(CAST({lhs} AS FLOAT), CAST({rhs} AS FLOAT))'
        return super().combine_expression(connector, sub_expressions)

    def convert_uuidfield_value(self, value, expression, connection):
        if value is not None:
            value = uuid.UUID(value)
        return value

    def format_for_duration_arithmetic(self, sql):
        if sql == 'NULL':
            return sql
        return 'CAST(%s AS BIGINT) MICROSECONDS' % sql

    def datetime_cast_date_sql(self, field_name, tzname):
        field_name = self._convert_field_to_tz(field_name, tzname)
        return "DATE(%s)" % field_name

    def field_cast_sql(self, db_type, internal_type):
        return super().field_cast_sql(db_type, internal_type)

    # Function to extract day, month or year from the date.
    # Reference: http://publib.boulder.ibm.com/infocenter/db2luw/v9r5/topic/com.ibm.db2.luw.sql.ref.doc/doc/r0023457.html
    def date_extract_sql(self, lookup_type, field_name):
        if lookup_type.upper() == 'WEEK_DAY':
            return " DAYOFWEEK(%s) " % (field_name)
        else:
            return " %s(%s) " % (lookup_type.upper(), field_name)

    def adapt_timefield_value(self, value):
        """
        Transform a time value to an object compatible with what is expected
        by the backend driver for time columns.
        """
        if value is None:
            return None

        # Expression values are adapted by the database.
        if hasattr(value, 'resolve_expression'):
            return value

        if timezone.is_aware(value):
            raise ValueError("Db2 for iSeries backend does not support timezone-aware times.")

        return value

    def adapt_datetimefield_value(self, value):
        """
        Transform a time value to an object compatible with what is expected
        by the backend driver for time columns.
        """
        if value is None:
            return None

        # Expression values are adapted by the database.
        if hasattr(value, 'resolve_expression'):
            return value

        if timezone.is_aware(value):
            if settings.USE_TZ:
                value = timezone.make_naive(value, self.connection.timezone)
            else:
                raise ValueError("Db2 for iSeries backend does not support timezone-aware times when USE_TZ is False")
        return value

    # Function to extract time zone-aware day, month or day of week from timestamps   
    def datetime_extract_sql(self, lookup_type, field_name, tzname):
        if settings.USE_TZ:
            raise NotImplemented

        if lookup_type.upper() == 'WEEK_DAY':
            return " DAYOFWEEK(%s) " % (field_name)
        else:
            return " %s(%s) " % (lookup_type.upper(), field_name)

    # Truncating the date value on the basic of lookup type.
    # e.g If input is 2008-12-04 and month then output will be 2008-12-01 00:00:00
    # Reference: http://www.ibm.com/developerworks/data/library/samples/db2/0205udfs/index.html
    def date_trunc_sql(self, lookup_type, field_name):
        sql = "TIMESTAMP(DATE(SUBSTR(CHAR(%s), 1, %d) || '%s'), TIME('00:00:00'))"
        if lookup_type.upper() == 'DAY':
            sql = sql % (field_name, 10, '')
        elif lookup_type.upper() == 'MONTH':
            sql = sql % (field_name, 7, '-01')
        elif lookup_type.upper() == 'YEAR':
            sql = sql % (field_name, 4, '-01-01')
        return sql

    # Truncating the time zone-aware timestamps value on the basic of lookup type
    def datetime_trunc_sql(self, lookup_type, field_name, tzname):
        sql = "TIMESTAMP(SUBSTR(CHAR(%s), 1, %d) || '%s')"
        if settings.USE_TZ:
            raise NotImplemented
        if lookup_type.upper() == 'SECOND':
            sql = sql % (field_name, 19, '.000000')
        if lookup_type.upper() == 'MINUTE':
            sql = sql % (field_name, 16, '.00.000000')
        elif lookup_type.upper() == 'HOUR':
            sql = sql % (field_name, 13, '.00.00.000000')
        elif lookup_type.upper() == 'DAY':
            sql = sql % (field_name, 10, '-00.00.00.000000')
        elif lookup_type.upper() == 'MONTH':
            sql = sql % (field_name, 7, '-01-00.00.00.000000')
        elif lookup_type.upper() == 'YEAR':
            sql = sql % (field_name, 4, '-01-01-00.00.00.000000')
        return sql, []

    def date_interval_sql(self, timedelta):
        return "%d DAYS + %d SECONDS + %d MICROSECONDS" % (
            timedelta.days, timedelta.seconds, timedelta.microseconds)

    # As casting is not required, so nothing is required to do in this function.
    def datetime_cast_sql(self):
        return "%s"

    # Function to return SQL from dropping foreign key.
    def drop_foreignkey_sql(self):
        return "DROP FOREIGN KEY"

    # Dropping auto generated property of the identity column.
    def drop_sequence_sql(self, table):
        return "ALTER TABLE %s ALTER COLUMN ID DROP IDENTITY" % (self.quote_name(table))

    def fulltext_search_sql(self, field_name):
        sql = "WHERE %s = ?" % field_name
        return sql

    # Function to return value of auto-generated field of last executed insert query. 
    def last_insert_id(self, cursor, table_name, pk_name):
        result = cursor.execute('SELECT IDENTITY_VAL_LOCAL() AS IDENTITY FROM SYSIBM.SYSDUMMY1')
        return result.fetchone()[0]

    # In case of WHERE clause, if the search is required to be case insensitive then converting 
    # left hand side field to upper.
    def lookup_cast(self, lookup_type, internal_type=None):
        if lookup_type in ('iexact', 'icontains', 'istartswith', 'iendswith'):
            return "UPPER(%s)"
        return "%s"

    # As DB2 v91 specifications, 
    # Maximum length of a table name and Maximum length of a column name is 128
    # http://publib.boulder.ibm.com/infocenter/db2e/v9r1/index.jsp?topic=/com.ibm.db2e.doc/db2elimits.html
    def max_name_length(self):
        return 128

    # As DB2 v97 specifications,
    # Maximum length of a database name is 8
    # http://publib.boulder.ibm.com/infocenter/db2luw/v9r7/topic/com.ibm.db2.luw.sql.ref.doc/doc/r0001029.html
    def max_db_name_length(self):
        return 8

    def no_limit_value(self):
        return None

    # Method to point custom query class implementation.
    def query_class(self, DefaultQueryClass):
        return query.query_class(DefaultQueryClass)

    # Function to quote the name of schema, table and column.
    def quote_name(self, name=None):
        name = name.upper()
        if (name.startswith("\"") & name.endswith("\"")):
            return name

        if (name.startswith("\"")):
            return "%s\"" % name

        if (name.endswith("\"")):
            return "\"%s" % name

        return "\"%s\"" % name

    # SQL to return RANDOM number.
    # Reference: http://publib.boulder.ibm.com/infocenter/db2luw/v8/topic/com.ibm.db2.udb.doc/admin/r0000840.htm
    def random_function_sql(self):
        return "SYSFUN.RAND()"

    def regex_lookup(self, lookup_type):
        if lookup_type == 'regex':
            return '''xmlcast( xmlquery('fn:matches(xs:string($c), "%%s")' passing %s as "c") as varchar(5)) = 'true' db2regexExtraField(%s)'''
        else:
            return '''xmlcast( xmlquery('fn:matches(xs:string($c), "%%s", "i")' passing %s as "c") as varchar(5)) = 'true' db2regexExtraField(%s)'''

    # As save-point is supported by DB2, following function will return SQL to create savepoint.
    def savepoint_create_sql(self, sid):
        return "SAVEPOINT %s ON ROLLBACK RETAIN CURSORS" % sid

    # Function to commit savepoint.   
    def savepoint_commit_sql(self, sid):
        return "RELEASE TO SAVEPOINT %s" % sid

    # Function to rollback savepoint.
    def savepoint_rollback_sql(self, sid):
        return "ROLLBACK TO SAVEPOINT %s" % sid

    # Deleting all the rows from the list of tables provided and resetting all the
    # sequences.
    def sql_flush(self, style, tables, sequences, allow_cascade=False):
        if tables:
            truncated_tables = set(t.upper() for t in tables)
            constraints = set()
            for table in tables:
                for foreign_table, constraint, foreign_pk_col, foreign_tgt_col in self._foreign_key_constraints(table):
                    if allow_cascade:
                        truncated_tables.add(foreign_table)
                    constraints.add((foreign_table, constraint, foreign_pk_col, foreign_tgt_col, table))

            sqls = [self._drop_constraint_sql(*constraint) for constraint in constraints]
            sqls.append('COMMIT')

            sqls.extend(' '.join([style.SQL_KEYWORD("DELETE"),
                                     style.SQL_KEYWORD("FROM"),
                                     style.SQL_TABLE("%s" % self.quote_name(table)),
                                     style.SQL_KEYWORD("WHERE"),
                                     "1 = 1"]) for table in tables)
            sqls.append('COMMIT')

            sqls.extend(' '.join([style.SQL_KEYWORD("ALTER TABLE"),
                                  style.SQL_TABLE("%s" % self.quote_name(sequence['table'])),
                                  style.SQL_KEYWORD("ALTER COLUMN"),
                                  self.quote_name(sequence['column']),
                                  style.SQL_KEYWORD("RESTART WITH 1")])
                        for sequence in sequences
                        if sequence['column'] is not None)

            sqls.extend(self._add_constraint_sql(*constraint) for constraint in constraints)
            return sqls
        else:
            return []

    def sequence_reset_sql(self, style, model_list):
        """
        Note: Db2 for iSeries cannot reset a table's primary key sequence if you have an open transaction
        that has already inserted/updated that table.
        """
        from django.db import models
        cursor = self.connection.cursor()
        sqls = []
        for model in model_list:
            table = model._meta.db_table
            for field in model._meta.local_fields:
                if isinstance(field, models.AutoField):
                    max_sql = "SELECT MAX(%s) FROM %s" % (self.quote_name(field.column), self.quote_name(table))
                    cursor.execute(max_sql)
                    max_id = [row[0] for row in cursor.fetchall()]
                    if max_id[0] == None:
                        max_id[0] = 0
                    sqls.append(style.SQL_KEYWORD("ALTER TABLE") + " " +
                                style.SQL_TABLE("%s" % self.quote_name(table)) +
                                " " + style.SQL_KEYWORD("ALTER COLUMN") + " %s "
                                % self.quote_name(field.column) +
                                style.SQL_KEYWORD("RESTART WITH %s" % (max_id[0] + 1)))
                    break

            for field in model._meta.many_to_many:
                m2m_table = field.m2m_db_table()
                if field.remote_field is not None and hasattr(field.remote_field, 'through'):
                    flag = field.remote_field.through
                else:
                    flag = False
                if not flag:
                    max_sql = "SELECT MAX(%s) FROM %s" % (self.quote_name('ID'), self.quote_name(table))
                    cursor.execute(max_sql)
                    max_id = [row[0] for row in cursor.fetchall()]
                    if max_id[0] == None:
                        max_id[0] = 0
                    sqls.append(style.SQL_KEYWORD("ALTER TABLE") + " " +
                                style.SQL_TABLE("%s" % self.quote_name(m2m_table)) +
                                " " + style.SQL_KEYWORD("ALTER COLUMN") + " %s "
                                % self.quote_name('ID') +
                                style.SQL_KEYWORD("RESTART WITH %s" % (max_id[0] + 1)))
        if cursor:
            cursor.close()

        return sqls

    # Returns sqls to reset the passed sequences
    def sequence_reset_by_name_sql(self, style, sequences):
        sqls = []
        for seq in sequences:
            sqls.append(style.SQL_KEYWORD("ALTER TABLE") + " " +
                        style.SQL_TABLE("%s" % self.quote_name(seq.get('table'))) +
                        " " + style.SQL_KEYWORD("ALTER COLUMN") + " %s " % self.quote_name(seq.get('column')) +
                        style.SQL_KEYWORD("RESTART WITH %s" % (1)))
        return sqls


    def value_to_db_datetime(self, value):
        if value is None:
            return None

        if is_aware(value):
            if settings.USE_TZ:
                value = value.astimezone(utc).replace(tzinfo=None)
            else:
                raise ValueError("Timezone aware datetime not supported")
        return str(value)

    def value_to_db_time(self, value):
        if value is None:
            return None

        if is_aware(value):
            raise ValueError("Timezone aware time not supported")
        else:
            return value

    def year_lookup_bounds_for_date_field(self, value):
        lower_bound = datetime.date(int(value), 1, 1)
        upper_bound = datetime.date(int(value), 12, 31)
        return [lower_bound, upper_bound]

    def bulk_insert_sql(self, fields, placeholder_rows):
        placeholder_rows_sql = (", ".join(row) for row in placeholder_rows)
        values_sql = ", ".join(f'({sql})' for sql in placeholder_rows_sql)
        return f"VALUES {values_sql}"

    def for_update_sql(self, nowait=False, skip_locked=False, of=()):
        """
        Return the FOR UPDATE SQL clause to lock rows for an update operation.
        """
        # DB2 doesn't support nowait select for update
        if nowait:
            raise utils.NotSupportedError("NOWAIT clause not supported by Db2 for iSeries")
        return 'FOR UPDATE%s%s' % (
            ' OF %s' % ', '.join(of) if of else '',
            ' SKIP LOCKED DATA' if skip_locked else '',
        )


    def fetch_returned_insert_id(self, cursor):
        return cursor.fetchone()[0]

    def fetch_returned_insert_ids(self, cursor):
        return [id_ for (id_, ) in cursor.fetchall()]

    def return_insert_id(self):
        """empty implementation as we implement returned ids with a cursor and custom Insert compiler"""
        return None, None

    def __foreign_key_constraints(self, table):
        foreign_keys_sql = """
        SELECT FK.TABLE_NAME, CST.CONSTRAINT_NAME, FK.COLUMN_NAME, TGT.COLUMN_NAME
            FROM QSYS2.SYSCST CST
                JOIN QSYS2.SYSKEYCST FK 
                  ON CST.CONSTRAINT_SCHEMA = FK.CONSTRAINT_SCHEMA 
                 AND CST.CONSTRAINT_NAME = FK.CONSTRAINT_NAME
                JOIN QSYS2.SYSREFCST REF 
                  ON CST.CONSTRAINT_SCHEMA = REF.CONSTRAINT_SCHEMA 
                 AND CST.CONSTRAINT_NAME = REF.CONSTRAINT_NAME
                JOIN QSYS2.SYSKEYCST PK 
                  ON REF.UNIQUE_CONSTRAINT_SCHEMA = PK.CONSTRAINT_SCHEMA 
                 AND REF.UNIQUE_CONSTRAINT_NAME = PK.CONSTRAINT_NAME
                JOIN QSYS2.SYSCOLUMNS TGT
                  ON PK.TABLE_SCHEMA = TGT.TABLE_SCHEMA
                 AND PK.TABLE_NAME = TGT.TABLE_NAME
                 AND TGT.ORDINAL_POSITION = FK.COLUMN_POSITION
            WHERE CST.CONSTRAINT_TYPE = 'FOREIGN KEY'
              AND FK.ORDINAL_POSITION = PK.ORDINAL_POSITION
              AND PK.TABLE_SCHEMA = CURRENT_SCHEMA
              AND PK.TABLE_NAME = ?
              AND ENABLED = 'YES'
          """.strip()
        cursor = self.connection.cursor()
        foreign_keys = cursor.execute(foreign_keys_sql, [table.upper()]).fetchall()
        return foreign_keys

    @cached_property
    def _foreign_key_constraints(self):
        """cached implementation, modelled after django's oracle backend implementation"""
        return lru_cache(maxsize=512)(self.__foreign_key_constraints)

    def _drop_constraint_sql(self, referencing_table, constraint_name, referencing_col, target_col, target_table):
        t = self.quote_name(referencing_table)
        c = self.quote_name(constraint_name)
        drop_constraint_sql = f"ALTER TABLE {t} DROP CONSTRAINT {c}"
        return drop_constraint_sql

    def _add_constraint_sql(self, referencing_table, constraint_name, referencing_col, target_col, target_table):
        ref_t = self.quote_name(referencing_table)
        cst = self.quote_name(constraint_name)
        col = self.quote_name(referencing_col)
        target_t = self.quote_name(target_table)
        add_constraint_sql = f"ALTER TABLE {ref_t} ADD CONSTRAINT {cst} " \
                             f"FOREIGN KEY ({col}) REFERENCES {target_t} ({target_col})"
        return add_constraint_sql

    def _convert_field_to_tz(self, field_name, tzname):
        # TODO: this function does not work properly on iseries
        if settings.USE_TZ:
            field_name = "CONVERT_TZ(%s, 'UTC', '%s')" % (field_name, tzname)
        return field_name

    def check_expression_support(self, expression):
        super().check_expression_support(expression)



"""Db2 for iSeries support for Django framework."""

__version__ = "1.0.3"

try:
    import pyodbc as Database
except:
    Database = None
else:
    pyodbc_version = tuple(int(x) for x in Database.version.split('.'))
    if pyodbc_version < (4, 0, 0):
        from django.core.exceptions import ImproperlyConfigured
        raise ImproperlyConfigured(f'PyODBC 4.0 or later required; you have {pyodbc_version}.')



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                      |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi              |
# +--------------------------------------------------------------------------+
from collections import namedtuple

from . import Database

try:
    from django.db.backends import BaseDatabaseIntrospection, FieldInfo
except ImportError:
    from django.db.backends.base.introspection import BaseDatabaseIntrospection, FieldInfo

TableInfo = namedtuple('TableInfo', ['name', 'type'])


class DatabaseIntrospection(BaseDatabaseIntrospection):
    """
    This is the class where database metadata information can be generated.
    """

    data_types_reverse = {
        'VARCHAR': 'CharField',
        'CHAR': 'CharField',
        'DATE': 'DateField',
        'TIMESTAMP': 'DateTimeField',
        'TIMESTMP': 'DateTimeField',
        'DECIMAL': 'DecimalField',
        'DOUBLE': 'FloatField',
        'INTEGER': 'IntegerField',
        'BIGINT': 'BigIntegerField',
        'SMALLINT': 'SmallIntegerField',
        'CLOB': 'TextField',
        'TIME': 'TimeField',
        'XML': 'XMLField',
        'BLOB': 'BinaryField',
    }

    def get_field_type(self, data_type, description):
        return super(DatabaseIntrospection, self).get_field_type(data_type, description)

    # Getting the list of all tables, which are present under current schema.
    def get_table_list(self, cursor):
        table_query = "SELECT TABLE_NAME, LOWER(TABLE_TYPE) FROM QSYS2.SYSTABLES WHERE TABLE_SCHEMA = CURRENT_SCHEMA"
        tables = cursor.execute(table_query)
        return [TableInfo(self.identifier_converter(t_name), t_type) for t_name, t_type in tables]

    # Generating a dictionary for foreign key details, which are present under current schema.
    def get_relations(self, cursor, table_name):
        relations = {}
        schema = cursor.get_current_schema()
        foreign_keys = list(cursor.foreignKeys(foreignTable=table_name.upper(), schema=schema))
        for fk in foreign_keys:
            relations[self.identifier_converter(fk.fkcolumn_name)] = (
                self.identifier_converter(fk.pktable_name),
                self.identifier_converter(fk.pkcolumn_name),
            )
        return relations

    # Private method. Getting Index position of column by its name
    def __get_col_index(self, cursor, schema, table_name, col_name):
        cols = cursor.columns(schema=schema, table=table_name, column=col_name)
        for col in cols:
            return col.ordinal_position - 1

    def get_key_columns(self, cursor, table_name):
        relations = []
        schema = cursor.get_current_schema()
        foreign_keys = list(cursor.foreignKeys(table=table_name.upper(), schema=schema))
        for fk in foreign_keys:
            relations.append((self.identifier_converter(fk.fkcolumn_name),
                              self.identifier_converter(fk.pktable_name),
                              self.identifier_converter(fk.pkcolumn_name)))
        return relations

    # Getting the description of the table.
    def get_table_description(self, cursor, table_name):
        qn = self.connection.ops.quote_name
        description = []

        sql = "SELECT TYPE FROM QSYS2.SYSTABLES WHERE TABLE_SCHEMA=CURRENT_SCHEMA AND TABLE_NAME=?"
        cursor.execute(sql, [table_name.upper()])
        table_type = cursor.fetchone()[0]

        if table_type != 'X':
            sql = """SELECT TRIM( COLUMN_NAME ) AS COLUMN_NAME
                          , TRIM( DATA_TYPE )   AS DATA_TYPE
                          , LENGTH              AS DISPLAY_SIZE
                          , STORAGE             AS INTERNAL_SIZE
                          , NUMERIC_PRECISION
                          , NUMERIC_SCALE
                          , CASE IS_NULLABLE WHEN 'Y' THEN 1 ELSE 0 END AS IS_NULLABLE
                     FROM QSYS2.SYSCOLUMNS C
                     WHERE TABLE_NAME = ?
                       AND TABLE_SCHEMA = CURRENT_SCHEMA
                     ORDER BY ORDINAL_POSITION"""
            column_descriptions = cursor.execute(sql, [table_name.upper()])

            for desc in column_descriptions:
                description.append(FieldInfo(
                    self.identifier_converter(desc[0]),
                    desc[1],
                    *desc[2:],
                    None
                ))
        return description

    def get_constraints(self, cursor, table_name):
        constraints = {}
        schema = cursor.get_current_schema()

        sql = """SELECT CST.CONSTRAINT_NAME
                     , COLUMN_NAME
                     , CASE CONSTRAINT_TYPE WHEN 'PRIMARY KEY' THEN 1 ELSE 0 END AS IS_PRIMARY
                     , CASE CONSTRAINT_TYPE WHEN 'UNIQUE' THEN 1 ELSE 0 END      AS IS_UNIQUE
                     , CASE CONSTRAINT_TYPE WHEN 'FOREIGN KEY' THEN 1 ELSE 0 END AS IS_FOREIGN
                     , CASE CONSTRAINT_TYPE WHEN 'CHECK' THEN 1 ELSE 0 END       AS IS_CHECK
                FROM QSYS2.SYSCST CST
                         JOIN QSYS2.SYSCSTCOL CSTCOL
                              ON ( CST.CONSTRAINT_NAME, CST.CONSTRAINT_SCHEMA ) = ( CSTCOL.CONSTRAINT_NAME, CSTCOL.CONSTRAINT_SCHEMA)
                WHERE CST.TABLE_NAME = ? and CST.TABLE_SCHEMA = CURRENT_SCHEMA"""
        cursor.execute(sql, [table_name.upper()])
        for constraint_name, column_name, is_primary, is_unique, is_foreign, is_check in cursor.fetchall():
            constraints.setdefault(self.identifier_converter(constraint_name), {
                'columns': [],
                'primary_key': is_primary,
                'unique': is_unique,
                'foreign_key': is_foreign,
                'check': is_check,
                'index': False
            })['columns'].append(self.identifier_converter(column_name))

        sql = ("SELECT IDX.INDEX_NAME, K.COLUMN_NAME, CASE IDX.IS_UNIQUE WHEN 'D' THEN 0 ELSE 1 END AS IS_UNIQUE "
               "  FROM QSYS2.SYSINDEXES IDX "
               "  JOIN QSYS2.SYSKEYS K ON ( IDX.INDEX_NAME, IDX.INDEX_SCHEMA ) = ( K.INDEX_NAME, K.INDEX_SCHEMA ) "
               " WHERE TABLE_NAME = ? "
               "   AND TABLE_SCHEMA = CURRENT_SCHEMA "
               "ORDER BY IDX.INDEX_NAME, K.ORDINAL_POSITION")
        indexes = cursor.execute(sql, [table_name.upper()])
        for index_name, column_name, is_unique in indexes.fetchall():
            constraints.setdefault(self.identifier_converter(index_name), {
                'columns': [],
                'primary_key': False,
                'unique': is_unique,
                'foreign_key': None,
                'check': False,
                'index': True,
            })['columns'].append(self.identifier_converter(column_name))
        return constraints

    def get_sequences(self, cursor, table_name, table_fields=()):
        from django.db import models

        seq_list = []
        for f in table_fields:
            if (isinstance(f, models.AutoField)):
                seq_list.append({'table': table_name, 'column': f.column})
                break
        return seq_list

    def identifier_converter(self, name):
        return name.lower()



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                      |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi              |
# +--------------------------------------------------------------------------+

"""
This module implements command line interface for DB2 through Django.
"""

try:
    from django.db.backends import BaseDatabaseClient
except ImportError:
    from django.db.backends.base.client import BaseDatabaseClient

import os


class DatabaseClient(BaseDatabaseClient):

    # Over-riding base method to provide shell support for DB2 through Django.
    def runshell(self):
        settings_dict = self.connection.settings_dict
        database_name = settings_dict['NAME']
        database_user = settings_dict['USER']
        database_password = settings_dict['PASSWORD']

        cmdArgs = ["db2"]

        if (os.name == 'nt'):
            cmdArgs += ["db2 connect to %s" % database_name]
        else:
            cmdArgs += ["connect to %s" % database_name]

        if (isinstance(database_user, str) and
                (database_user != '')):
            cmdArgs += ["user %s" % database_user]

            if (isinstance(database_password, str) and
                    (database_password != '')):
                cmdArgs += ["using %s" % database_password]

        # db2cmd is the shell which is required to run db2 commands on windows.
        if (os.name == 'nt'):
            os.execvp('db2cmd', cmdArgs)
        else:
            os.execvp('db2', cmdArgs)



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                      |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi              |
# +--------------------------------------------------------------------------+

"""
Custom Query class for DB2.
Derives from: django.db.models.sql.query.Query
"""


def query_class(QueryClass):
    class DB2QueryClass(QueryClass):
        __rownum = 'Z.__ROWNUM'

        # To get ride of LIMIT/OFFSET problem in DB2, this method has been implemented.
        def as_sql(self, with_limits=True, with_col_aliases=False):
            if not (with_limits and (self.high_mark is not None or self.low_mark)):
                return super(DB2QueryClass, self).as_sql(False, with_col_aliases)
            else:
                if self.high_mark == self.low_mark:
                    return '', ()

                sql_ori, params = super(DB2QueryClass, self).as_sql(False, with_col_aliases)
                sql_split = sql_ori.split(" FROM ")

                sql_sec = ""
                if len(sql_split) > 2:
                    for i in range(1, len(sql_split)):
                        sql_sec = " %s FROM %s " % (sql_sec, sql_split[i])
                else:
                    sql_sec = " FROM %s " % (sql_split[1])

                dummyVal = "Z.__db2_"
                sql_pri = ""

                sql_sel = "SELECT"
                if self.distinct:
                    sql_sel = "SELECT DISTINCT"

                sql_select_token = sql_split[0].split(",")
                i = 0
                while (i < len(sql_select_token)):
                    if sql_select_token[i].count("TIMESTAMP(DATE(SUBSTR(CHAR(") == 1:
                        sql_sel = "%s \"%s%d\"," % (sql_sel, dummyVal, i + 1)
                        sql_pri = '%s %s,%s,%s,%s AS "%s%d",' % (
                            sql_pri,
                            sql_select_token[i],
                            sql_select_token[i + 1],
                            sql_select_token[i + 2],
                            sql_select_token[i + 3],
                            dummyVal, i + 1)
                        i = i + 4
                        continue

                    if sql_select_token[i].count(" AS ") == 1:
                        temp_col_alias = sql_select_token[i].split(" AS ")
                        sql_pri = '%s %s,' % (sql_pri, sql_select_token[i])
                        sql_sel = "%s %s," % (sql_sel, temp_col_alias[1])
                        i = i + 1
                        continue

                    sql_pri = '%s %s AS "%s%d",' % (sql_pri, sql_select_token[i], dummyVal, i + 1)
                    sql_sel = "%s \"%s%d\"," % (sql_sel, dummyVal, i + 1)
                    i = i + 1

                sql_pri = sql_pri[:len(sql_pri) - 1]
                sql_pri = "%s%s" % (sql_pri, sql_sec)
                sql_sel = sql_sel[:len(sql_sel) - 1]
                sql = '%s, ( ROW_NUMBER() OVER() ) AS "%s" FROM ( %s ) AS M' % (sql_sel, self.__rownum, sql_pri)
                sql = '%s FROM ( %s ) Z WHERE' % (sql_sel, sql)

                if self.low_mark != 0:
                    sql = '%s "%s" > %d' % (sql, self.__rownum, self.low_mark)

                if self.low_mark != 0 and self.high_mark is not None:
                    sql = '%s AND ' % (sql)

                if self.high_mark is not None:
                    sql = '%s "%s" <= %d' % (sql, self.__rownum, self.high_mark)

            return sql, params

        # http://www.python.org/dev/peps/pep-0307/
        # See Extended __reduce__ API
        def __reduce__(self):
            return (__newobj__, (QueryClass,))

        # For case insensitive search, converting parameter value to upper case.
        # The right hand side will get converted to upper case in the SQL itself.
        from django.db.models.sql.where import AND
        def add_filter(self, filter_expr, connector=AND, negate=False, trim=False,
                       can_reuse=None, process_extras=True):
            if len(filter_expr) != 0 and filter_expr is not None:
                filter_expr = list(filter_expr)
                if filter_expr[0].find("__iendswith") != -1 or \
                        filter_expr[0].find("__istartswith") != -1 or \
                        filter_expr[0].find("__icontains") != -1 or \
                        filter_expr[0].find("__iexact") != -1:
                    filter_expr[1] = filter_expr[1].upper()

                filter_expr = tuple(filter_expr)
            return super(DB2QueryClass, self).add_filter(filter_expr, connector,
                                                         negate, trim, can_reuse, process_extras)

    return DB2QueryClass


# Method to make DB2QueryClass picklable
def __newobj__(QueryClass):
    # http://www.python.org/dev/peps/pep-0307/
    # The __newobj__ unpickling function
    DB2QueryClass = query_class(QueryClass)
    return DB2QueryClass.__new__(DB2QueryClass)



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                      |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi              |
# +--------------------------------------------------------------------------+



import datetime
import platform
import re

# For checking django's version
from functools import partial
from typing import Optional

import sqlparse
from django.db import utils

from django_iseries import Database

dbms_name = 'dbms_name'

DatabaseError = Database.DatabaseError
IntegrityError = Database.IntegrityError
Error = Database.Error
InterfaceError = Database.InterfaceError
DataError = Database.DataError
OperationalError = Database.OperationalError
InternalError = Database.InternalError
ProgrammingError = Database.ProgrammingError
NotSupportedError = Database.NotSupportedError


FORMAT_QMARK_REGEX = re.compile(r'(?<!%)%s')
SQLCODE_0530_REGEX = re.compile("^(\[.+\] *){4}SQL0530.*")
SQLCODE_0910_REGEX = re.compile("^(\[.+\] *){4}SQL0910.*")


class DatabaseWrapper:
    # Get new database connection for non persistance connection 
    def get_new_connection(self, kwargs):
        driver_name = 'iSeries Access ODBC Driver' if platform.system() == 'Windows' else 'IBM i Access ODBC Driver'
        if 'port' in kwargs and 'host' in kwargs:
            kwargs['dsn'] = f"DRIVER={{{driver_name}}};DATABASE=%s;UNICODESQL=1;XDYNAMIC=1;" \
                            f"PKG=A/DJANGO,2,0,0,1,512;" \
                            "SYSTEM=%s;PORT=%s;PROTOCOL=TCPIP;UID=%s;PWD=%s" % (
                kwargs.get('database'),
                kwargs.get('host'),
                kwargs.get('port'),
                kwargs.get('user'),
                kwargs.get('password')
            )
        else:
            kwargs['dsn'] = kwargs.get('database')

        if 'security' in kwargs:
            kwargs['dsn'] += "security=%s;" % (kwargs.get('security'))
            del kwargs['security']

        if 'sslclientkeystoredb' in kwargs:
            kwargs['dsn'] += "SSLCLIENTKEYSTOREDB=%s;" % (kwargs.get('sslclientkeystoredb'))
            del kwargs['sslclientkeystoredb']

        if 'sslclientkeystoredbpassword' in kwargs:
            kwargs['dsn'] += "SSLCLIENTKEYSTOREDBPASSWORD=%s;" % (kwargs.get('sslclientkeystoredbpassword'))
            del kwargs['sslclientkeystoredbpassword']

        if 'sslclientkeystash' in kwargs:
            kwargs['dsn'] += "SSLCLIENTKEYSTASH=%s;" % (kwargs.get('sslclientkeystash'))
            del kwargs['sslclientkeystash']

        if 'sslservercertificate' in kwargs:
            kwargs['dsn'] += "SSLSERVERCERTIFICATE=%s;" % (kwargs.get('sslservercertificate'))
            del kwargs['sslservercertificate']

        conn_options = {'autocommit': False}
        kwargs['conn_options'] = conn_options
        if 'options' in kwargs:
            kwargs.update(kwargs.get('options'))
            del kwargs['options']
        if 'port' in kwargs:
            del kwargs['port']

        currentschema = kwargs.pop('currentschema', None)

        dsn = kwargs.pop('dsn', '')

        connection = Database.connect(dsn, **kwargs)
        if currentschema:
            cursor = DB2CursorWrapper(connection)
            cursor.set_current_schema(currentschema)

        return connection

    def is_active(self, connection=None):
        return Database.ibm_db.active(connection.conn_handler)

    # Over-riding _cursor method to return DB2 cursor.
    def _cursor(self, connection):
        return DB2CursorWrapper(connection)

    def close(self, connection):
        try:
            connection.close()
        except ProgrammingError as e:
            if str(e) == 'Attempt to use a closed connection.':
                pass
            else:
                raise

    def get_server_version(self, connection):
        self.connection = connection
        if not self.connection:
            self.cursor()
        return tuple(int(version) for version in self.connection.server_info()[1].split("."))


class DB2CursorWrapper:
    """
    This is the wrapper around IBM_DB_DBI in order to support format parameter style
    IBM_DB_DBI supports qmark, where as Django support format style, 
    hence this conversion is required.

    pyodbc.Cursor cannot be subclassed, so we store it as an attribute
    """

    def __init__(self, connection):
        self.cursor: Database.Cursor = connection.cursor()

    def __iter__(self):
        return self.cursor

    def __getattr__(self, attr):
        return getattr(self.cursor, attr)

    def get_current_schema(self):
        self.execute('select CURRENT_SCHEMA from sysibm.sysdummy1')
        return self.fetchone()[0]

    def set_current_schema(self, schema):
        self.execute(f'SET CURRENT_SCHEMA = {schema}')

    def close(self):
        """
        Django calls close() twice on some cursors, but pyodbc does not allow this.
        pyodbc deletes the 'connection' attribute when closing a cursor, so we check for that.

        In the unlikely event that this code prevents close() from being called, pyodbc will close
        the cursor automatically when it goes out of scope.
        """
        if getattr(self, 'connection', False):
            self.cursor.close()

    def execute(self, query, params=()):
        if params:
            query = self.convert_query(query)
            query, params = self._replace_placeholders_in_select_clause(params, query)
        result = self._wrap_execute(partial(self.cursor.execute, query, params))
        return result

    def executemany(self, query, param_list):
        if not param_list:
            # empty param_list means do nothing (execute the query zero times)
            return
        query = self.convert_query(query)
        result = self._wrap_execute(partial(self.cursor.executemany, query, param_list))
        return result

    def _wrap_execute(self, execute):
        try:
            result = execute()
        except Database.Error as e:
            # iaccess seems to be sending incorrect sqlstate for some errors
            # reraise "referential constraint violation" errors as IntegrityError
            if e.args[0] == 'HY000' and SQLCODE_0530_REGEX.match(e.args[1]):
                raise utils.IntegrityError(*e.args, execute.func, *execute.args)
            elif e.args[0] == 'HY000' and SQLCODE_0910_REGEX.match(e.args[1]):
                # file in use error (likely in the same transaction)
                query, params, *_ = execute.args
                if query.startswith('ALTER TABLE') and 'RESTART WITH' in query:
                    raise utils.ProgrammingError(
                        *e.args,
                        execute.func,
                        execute.args,
                        "Db2 for iSeries cannot reset a table's primary key sequence during same "
                        "transaction as insert/update on that table"
                    )
            raise type(e)(*e.args, execute.func, execute.args)
        if result == self.cursor:
            return self
        return result

    def _replace_placeholders_in_select_clause(self, params, query):
        """Db2 for i does not allow placeholders in select clause; this converts them to literals"""
        if isinstance(params, tuple):
            params = list(params)
        current_param_idx = -1
        in_select_clause = False
        tmp = []
        parsed_statement = sqlparse.parse(query)[0]
        for t in parsed_statement.flatten():
            if t.ttype == sqlparse.tokens.Name.Placeholder:  # '?'
                current_param_idx += 1
                if in_select_clause:
                    quoted_val = self.quote_value(params.pop(current_param_idx))
                    t = sqlparse.sql.Token(sqlparse.tokens.String, quoted_val)
            elif t.normalized == 'SELECT':
                in_select_clause = True
            elif t.normalized == 'FROM':
                in_select_clause = False
            tmp.append(str(t))
        query = ''.join(tmp)
        return query, params

    def convert_query(self, query):
        """
        Django uses "format" style placeholders, but the iaccess odbc driver uses "qmark" style.
        This fixes it -- but note that if you want to use a literal "%s" in a query,
        you'll need to use "%%s".
        """
        return FORMAT_QMARK_REGEX.sub('?', query).replace('%%', '%')

    def _row_factory(self, row: Optional[Database.Row]):
        if row is None:
            return row
        return tuple(row)

    def fetchone(self):
        return self._row_factory(self.cursor.fetchone())

    def fetchmany(self, size):
        return [self._row_factory(row) for row in self.cursor.fetchmany(size)]

    def fetchall(self):
        return [self._row_factory(row) for row in self.cursor.fetchall()]

    @property
    def last_identity_val(self):
        result = self.execute('SELECT IDENTITY_VAL_LOCAL() AS IDENTITY FROM SYSIBM.SYSDUMMY1')
        row = result.fetchone()
        return row[0]

    def quote_value(self, value):
        if isinstance(value, (datetime.datetime, datetime.date, datetime.time, str)):
            return f"'{value}'"
        if isinstance(value, bool):
            return '1' if value else '0'
        return str(value)



import os
import random
import string

SECRET_KEY = [random.choice(string.ascii_lowercase) for i in range(10)]

DATABASES = {
    'default':
        {
            'ENGINE': 'django_iseries',
            'NAME': 'iseries',  # arbitrary database name for db2 on iseries
            'HOST': os.environ['TEST_SYSTEM_HOST'],
            'USER': os.environ['TEST_SYSTEM_USERNAME'],
            'PASSWORD': os.environ['TEST_SYSTEM_PASSWORD'],
            'CURRENTSCHEMA': os.environ['TEST_SYSTEM_SCHEMA'],
        },
    'other':
        {
            'ENGINE': 'django.db.backends.sqlite3',
        }
}

# Use a fast hasher to speed up tests.
PASSWORD_HASHERS = [
    'django.contrib.auth.hashers.MD5PasswordHasher',
]

INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.sites',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.admin.apps.SimpleAdminConfig',
    'django.contrib.staticfiles',

    'tests',
]

MIGRATION_MODULES = {
    # This lets us skip creating migrations for the test models as many of
    # them depend on one of the following contrib applications.
    'auth': None,
    'contenttypes': None,
    'sessions': None,
}



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                 |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi              |
# +--------------------------------------------------------------------------+

from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.creation import BaseDatabaseCreation
from django.db.backends.utils import truncate_name

TEST_DBNAME_PREFIX = 'test_'


class DatabaseCreation(BaseDatabaseCreation):
    psudo_column_prefix = 'psudo_'

    data_types = {
        'AutoField': 'INTEGER GENERATED BY DEFAULT AS IDENTITY '
                     '(START WITH 1, INCREMENT BY 1, CACHE 10 ORDER)',  # DB2 Specific
        'BigAutoField': 'BIGINT GENERATED BY DEFAULT AS IDENTITY '
                        '(START WITH 1, INCREMENT BY 1, CACHE 10 ORDER)',  # DB2 Specific
        'CharField': 'VARCHAR(%(max_length)s) CCSID 1208',
        'CommaSeparatedIntegerField': 'VARCHAR(%(max_length)s)',
        'DateField': 'DATE',
        'DateTimeField': 'TIMESTAMP',
        'DecimalField': 'DECIMAL(%(max_digits)s, %(decimal_places)s)',
        'FileField': 'VARCHAR(%(max_length)s)',
        'FilePathField': 'VARCHAR(%(max_length)s)',
        'FloatField': 'DOUBLE',
        'ImageField': 'VARCHAR(%(max_length)s)',
        'IntegerField': 'INTEGER',
        'BigIntegerField': 'BIGINT',
        'IPAddressField': 'VARCHAR(15)',
        'GenericIPAddressField': 'VARCHAR(39)',
        'ManyToManyField': 'VARCHAR(%(max_length)s)',
        'OneToOneField': 'VARCHAR(%(max_length)s)',
        'PhoneNumberField': 'VARCHAR(16)',
        'SlugField': 'VARCHAR(%(max_length)s)',
        'SmallIntegerField': 'SMALLINT',
        'TextField': 'CLOB CCSID 1208',
        'TimeField': 'TIME',
        'USStateField': 'VARCHAR(2)',
        'URLField': 'VARCHAR2(%(max_length)s)',
        'XMLField': 'XML',
        'BinaryField': 'BLOB',
        'UUIDField': 'VARCHAR(255)',
        "DurationField": 'DOUBLE',
        'BooleanField': 'SMALLINT',
        'NullBooleanField': 'SMALLINT',
        'PositiveIntegerField': 'INTEGER',
        'PositiveSmallIntegerField': 'SMALLINT',
    }

    data_type_check_constraints = {
        'BooleanField': '%(attname)s IN (0,1)',
        'NullBooleanField': '(%(attname)s IN (0,1)) OR (%(attname)s IS NULL)',
        'PositiveIntegerField': '%(attname)s >= 0',
        'PositiveSmallIntegerField': '%(attname)s >= 0',
    }

    def sql_indexes_for_field(self, model, f, style):
        """Return the CREATE INDEX SQL statements for a single model field"""
        output = []
        qn = self.connection.ops.quote_name
        max_name_length = self.connection.ops.max_name_length()
        i = 0
        if len(model._meta.unique_together_index) != 0:
            for unique_together_index in model._meta.unique_together_index:
                i = i + 1
                column_list = []
                for column in unique_together_index:
                    for local_field in model._meta.local_fields:
                        if column == local_field.name:
                            column_list.extend([local_field.column])

                self.__add_psudokey_column(style, self.connection.cursor(), model._meta.db_table,
                                           model._meta.pk.attname, column_list)
                column_list.extend(
                    [truncate_name("%s%s" % (self.psudo_column_prefix, "_".join(column_list)), max_name_length)])
                output.extend([style.SQL_KEYWORD('CREATE UNIQUE INDEX') + ' ' + \
                               style.SQL_TABLE(qn('db2_%s_%s' % (model._meta.db_table, i))) + ' ' + \
                               style.SQL_KEYWORD('ON') + ' ' + \
                               style.SQL_TABLE(qn(model._meta.db_table)) + ' ' + \
                               '( %s )' % ", ".join(column_list)])
            model._meta.unique_together_index = []

        if f.unique_index:
            column_list = []
            column_list.extend([f.column])
            self.__add_psudokey_column(style, self.connection.cursor(), model._meta.db_table,
                                       model._meta.pk.attname, column_list)
            cisql = 'CREATE UNIQUE INDEX'
            output.extend([style.SQL_KEYWORD(cisql) + ' ' +
                           style.SQL_TABLE(qn('%s_%s' % (model._meta.db_table, f.column))) + ' ' +
                           style.SQL_KEYWORD('ON') + ' ' +
                           style.SQL_TABLE(qn(model._meta.db_table)) + ' ' +
                           "(%s, %s )" % (style.SQL_FIELD(qn(f.column)), style.SQL_FIELD(
                qn(truncate_name((self.psudo_column_prefix + f.column), max_name_length))))])
            return output

        if f.db_index and not f.unique:
            cisql = 'CREATE INDEX'
            output.extend([style.SQL_KEYWORD(cisql) + ' ' +
                           style.SQL_TABLE(qn('%s_%s' % (model._meta.db_table, f.column))) + ' ' +
                           style.SQL_KEYWORD('ON') + ' ' +
                           style.SQL_TABLE(qn(model._meta.db_table)) + ' ' +
                           "(%s)" % style.SQL_FIELD(qn(f.column))])

        return output

    def _create_test_db(self, verbosity, autoclobber, keepdb=False):
        if keepdb:
            return
        else:
            raise ImproperlyConfigured('pyodbc iseries driver does not support database creation')

    def _destroy_test_db(self, test_database_name, verbosity):
        raise ImproperlyConfigured('pyodbc iseries driver does not support database destruction (use --keepdb)')

    # Private method to clean up database.
    def __clean_up(self, cursor):
        tables = self.connection.introspection.django_table_names(only_existing=True)
        for table in tables:
            sql = "DROP TABLE %s" % self.connection.ops.quote_name(table)
            cursor.execute(sql)
        cursor.close()

    # Private method to alter a table with adding psudokey column
    def __add_psudokey_column(self, style, cursor, table_name, pk_name, column_list):
        qn = self.connection.ops.quote_name
        max_name_length = self.connection.ops.max_name_length()

        sql = style.SQL_KEYWORD('ALTER TABLE ') + \
              style.SQL_TABLE(qn(table_name)) + \
              style.SQL_KEYWORD(' ADD COLUMN ') + \
              style.SQL_FIELD(
                  qn(truncate_name("%s%s" % (self.psudo_column_prefix, "_".join(column_list)), max_name_length))) + \
              style.SQL_KEYWORD(' GENERATED ALWAYS AS( CASE WHEN ') + \
              style.SQL_FIELD("%s %s" % (" IS NULL OR ".join(column_list), 'IS NULL THEN ')) + \
              style.SQL_FIELD(qn(pk_name)) + \
              style.SQL_KEYWORD(' END ) ;')
        cursor.execute('SET INTEGRITY FOR ' + style.SQL_TABLE(qn(table_name)) + ' OFF CASCADE DEFERRED;')
        cursor.execute(sql)
        cursor.execute('SET INTEGRITY FOR ' + style.SQL_TABLE(table_name) + ' IMMEDIATE CHECKED;')
        cursor.close()

    # private method to create dictionary of login credentials for test database
    def __create_test_kwargs(self):
        if (isinstance(self.connection.settings_dict['NAME'], str) and
                (self.connection.settings_dict['NAME'] != '')):
            database = self.connection.settings_dict['NAME']
        else:
            from django.core.exceptions import ImproperlyConfigured
            raise ImproperlyConfigured("the database Name doesn't exist")
        database_user = self.connection.settings_dict['USER']
        database_pass = self.connection.settings_dict['PASSWORD']
        database_host = self.connection.settings_dict['HOST']
        database_port = self.connection.settings_dict['PORT']
        self.connection.settings_dict['SUPPORTS_TRANSACTIONS'] = True

        kwargs = {}
        kwargs['database'] = database
        if isinstance(database_user, str):
            kwargs['user'] = database_user

        if isinstance(database_pass, str):
            kwargs['password'] = database_pass

        if isinstance(database_host, str):
            kwargs['host'] = database_host

        if isinstance(database_port, str):
            kwargs['port'] = database_port

        if isinstance(database_host, str):
            kwargs['host'] = database_host

        return kwargs



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                 |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Rahul Priyadarshi, Hemlata Bhatt, Vyshakh A                     |
# +--------------------------------------------------------------------------+

import copy
import datetime
import uuid

try:
    from django.db.backends.schema import BaseDatabaseSchemaEditor
except ImportError:
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor

from django.db import models
from django.db.backends.utils import truncate_name
from django.db.models.fields.related import ManyToManyField
from django import VERSION as djangoVersion
from . import Database

Error = Database.Error


class DB2SchemaEditor(BaseDatabaseSchemaEditor):
    psudo_column_prefix = 'psudo_'
    sql_delete_table = "DROP TABLE %(table)s"
    sql_rename_table = "RENAME TABLE %(old_table)s TO %(new_table)s"
    sql_create_column = "ALTER TABLE %(table)s ADD COLUMN %(column)s %(definition)s"
    sql_alter_column_type = "ALTER COLUMN %(column)s SET DATA TYPE %(type)s"
    sql_alter_column_type_from_int_to_auto = "ALTER COLUMN %(column)s SET GENERATED BY DEFAULT AS IDENTITY( START WITH %(max)d, INCREMENT BY 1, CACHE 10 ORDER )"
    sql_create_fk = "ALTER TABLE %(table)s ADD CONSTRAINT %(name)s FOREIGN KEY (%(column)s) REFERENCES %(to_table)s (%(to_column)s)"
    sql_delete_pk = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s"
    sql_delete_unique = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s"
    sql_drop_pk = "ALTER TABLE %(table)s DROP PRIMARY KEY"
    sql_drop_default = "ALTER TABLE %(table)s ALTER COLUMN %(column)s DROP DEFAULT"

    @property
    def sql_create_pk(self):
        self._reorg_tables()
        return "ALTER TABLE %(table)s ADD CONSTRAINT %(name)s PRIMARY KEY (%(columns)s)"

    def prepare_default(self, value):
        return self.quote_value(value)

    def alter_field(self, model, old_field, new_field, strict=False):
        alter_field_data_type = False
        alter_field_nullable = False
        alter_field_default = False
        alter_field_primary_key = False
        alter_field_name = False
        alter_field_check_constraint = False
        alter_field_unique = False
        alter_field_index = False
        rebuild_incomming_fk = False
        alter_incomming_fk_data_type = False
        deferred_constraints = {
            'pk': {},
            'unique': {},
            'index': {},
            'check': {}
        }

        old_db_field = old_field.db_parameters(connection=self.connection)
        new_db_field = new_field.db_parameters(connection=self.connection)
        old_db_field_type = old_db_field['type']
        new_db_field_type = new_db_field['type']

        if old_field.remote_field is not None and hasattr(old_field.remote_field, 'through'):
            rel_condition = (
                        old_field.remote_field.through and new_field.remote_field.through and old_field.remote_field.through._meta.auto_created and new_field.remote_field.through._meta.auto_created)
        else:
            rel_condition = False

        if ((old_db_field_type, new_db_field_type) == (None, None)) and rel_condition:
            return self._alter_many_to_many(model, old_field, new_field, strict)
        elif old_db_field_type is None or new_db_field_type is None:
            raise ValueError("Cannot alter field %s into %s" % (
                old_db_field,
                new_db_field,
            ))

        # Need to change datatype which need remaking of field
        if (old_db_field_type != new_db_field_type) and (
                isinstance(old_field, (models.AutoField, models.TextField)) or isinstance(new_field, models.TextField)):
            if old_field.primary_key and new_field.primary_key:
                rebuild_incomming_fk = True
                if isinstance(old_field, models.AutoField) and isinstance(new_field, models.IntegerField):
                    pass
                else:
                    alter_incomming_fk_data_type = True
            old_field, new_field = self.alterFieldDataTypeByRemaking(model, old_field, new_field, strict)
            old_db_field = old_field.db_parameters(connection=self.connection)
            new_db_field = new_field.db_parameters(connection=self.connection)
            old_db_field_type = old_db_field['type']
            new_db_field_type = new_db_field['type']

        if old_db_field_type != new_db_field_type:
            alter_field_data_type = True

        if old_field.column != new_field.column:
            alter_field_name = True
        if old_field.db_index != new_field.db_index:
            alter_field_index = True
        if old_field.unique != new_field.unique:
            alter_field_unique = True
        if old_field.primary_key != new_field.primary_key:
            alter_field_primary_key = True
        if old_db_field['check'] != new_db_field['check']:
            alter_field_check_constraint = True
        if old_field.null != new_field.null:
            alter_field_nullable = True

        old_default = self.effective_default(old_field)
        new_default = self.effective_default(new_field)
        if (old_field.default is not None) and old_field.has_default():
            if old_default != new_default:
                alter_field_default = True

        # Need to remove Primary Key
        if alter_field_primary_key and old_field.primary_key:
            if strict:
                pk_names = self._constraint_names(model, [old_field.column], primary_key=True)
                if len(pk_names) == 0:
                    raise ValueError("Found no primary key in %s.%s " % (model._meta.db_table, old_field.column))
            self.execute(
                self.sql_drop_pk % {
                    'table': self.quote_name(model._meta.db_table)
                }
            )

        # Need to remove unique Key
        if alter_field_unique and old_field.unique or (
                old_field.unique and alter_field_primary_key and not old_field.primary_key):
            unique_key_names = self._constraint_names(model, [old_field.column], unique=True)
            if strict and len(unique_key_names) != 1:
                raise ValueError("Found wrong number of unique constraints for (table)s.(column)s" % {
                    'table': model._meta.db_table, 'column': old_field.column
                })

            for unique_key_name in unique_key_names:
                self.execute(
                    self.sql_delete_unique % {
                        'table': self.quote_name(model._meta.db_table),
                        'name': unique_key_name
                    }
                )

        # Need to remove Index
        if alter_field_index and old_field.db_index:
            index_names = self._constraint_names(model, [old_field.column], index=True)
            if strict and len(index_names) != 1:
                raise ValueError("Found wrong number of Indexes for (table)s.(column)s" % {
                    'table': model._meta.db_table, 'column': old_field.column
                })
            for index_name in index_names:
                self.execute(
                    self.sql_delete_index % {
                        'name': index_name
                    }
                )

        # Need to remove check constraint
        if alter_field_check_constraint and old_db_field['check']:
            check_constraint_names = self._constraint_names(model, [old_field.column], check=True)
            if strict and len(check_constraint_names) != 1:
                raise ValueError("Found wrong number of check constraints for (table)s.(column)s" % {
                    'table': model._meta.db_table, 'column': old_field.column
                })
            for check_constraint_name in check_constraint_names:
                self.execute(
                    self.sql_delete_check % {
                        'table': self.quote_name(model._meta.db_table),
                        'name': check_constraint_name
                    }
                )

        # Need to remove Nullability
        if alter_field_nullable and old_field.null:
            sql = self.sql_alter_column_not_null % {'column': self.quote_name(old_field.column)}
            self.execute(
                self.sql_alter_column % {
                    'table': self.quote_name(model._meta.db_table),
                    'changes': sql
                }
            )

        # Drop all FK constraints, if require we will make it again
        flag = old_field.remote_field
        if flag:
            fk_names = self._constraint_names(model, [old_field.column], foreign_key=True)
            for fk_name in fk_names:
                self.execute(
                    self.sql_delete_fk % {
                        'table': self.quote_name(model._meta.db_table),
                        'name': fk_name
                    }
                )

        if alter_field_name or alter_field_data_type:

            # Drop all incoming FK constraint, if require we will make it again
            if old_field.primary_key and new_field.primary_key:
                rebuild_incomming_fk = True
                for incoming_fks in old_field.model._meta.get_fields():
                    fk_names = self._constraint_names(incoming_fks.model, [incoming_fks.field.column], foreign_key=True)
                    for fk_name in fk_names:
                        self.execute(
                            self.sql_delete_fk % {
                                'table': self.quote_name(incoming_fks.model._meta.db_table),
                                'name': fk_name,
                            }
                        )

            # Defer constraint check
            with self.connection.cursor() as cur:
                constraints = self.connection.introspection.get_constraints(cur, model._meta.db_table)
            self._defer_constraints_check(constraints, deferred_constraints, old_field, new_field, model, defer_pk=True,
                                          defer_unique=True, defer_index=True, defer_check=True)

            # Need to change the field name
            if alter_field_name:
                self.execute(
                    self.sql_rename_column % {
                        'table': self.quote_name(model._meta.db_table),
                        'old_column': self.quote_name(old_field.column),
                        'new_column': self.quote_name(new_field.column),
                    }
                )

            # Need to change the field type
            if alter_field_data_type:
                if old_field.primary_key and new_field.primary_key:
                    if isinstance(new_field, models.AutoField) and isinstance(old_field, models.IntegerField):
                        pass
                    else:
                        alter_incomming_fk_data_type = True
                # Will make default later
                if (old_field.default is not None) and (old_field.has_default()) and (old_default is not None):
                    self.execute(self.sql_drop_default % {
                        'table': self.quote_name(model._meta.db_table),
                        'column': self.quote_name(new_field.column)
                    }
                                 )
                if isinstance(new_field, models.AutoField):
                    with self.connection.cursor() as cur:
                        cur.execute(
                            'SELECT MAX( %(column)s ) from %(table)s' % {
                                'column': self.quote_name(new_field.column),
                                'table': self.quote_name(model._meta.db_table),
                            }
                        )
                        max = cur.fetchone()
                        if max[0] is None:
                            max = 0
                        else:
                            max = max[0]
                    if not isinstance(old_field, models.IntegerField):
                        sql = self.sql_alter_column_type % {
                            'column': self.quote_name(new_field.column),
                            'type': 'Integer'
                        }
                        self.execute(
                            self.sql_alter_column % {
                                'table': self.quote_name(model._meta.db_table),
                                'changes': sql
                            }
                        )
                    sql = self.sql_alter_column_type_from_int_to_auto % {
                        'column': self.quote_name(new_field.column),
                        'max': max + 1
                    }
                    self.execute(
                        self.sql_alter_column % {
                            'table': self.quote_name(model._meta.db_table),
                            'changes': sql
                        }
                    )
                else:
                    sql = self.sql_alter_column_type % {
                        'column': self.quote_name(new_field.column),
                        'type': new_db_field_type
                    }
                    self.execute(
                        self.sql_alter_column % {
                            'table': self.quote_name(model._meta.db_table),
                            'changes': sql
                        }
                    )

            # restore constraint checks
            self._restore_constraints_check(deferred_constraints, old_field, new_field, model)

        # Need to change Default
        if alter_field_default:
            if new_default is None:
                if alter_field_data_type or alter_field_nullable:
                    pass
                else:
                    self.execute(self.sql_drop_default % {
                        'table': self.quote_name(model._meta.db_table),
                        'column': self.quote_name(new_field.column)
                    }
                                 )
            else:
                sql = self.sql_alter_column_default % {
                    'column': self.quote_name(new_field.column),
                    'default': self.prepare_default(new_default),
                }
                self.execute(
                    self.sql_alter_column % {
                        'table': self.quote_name(model._meta.db_table),
                        'changes': sql
                    }
                )

        # Need to change nullability
        if alter_field_nullable:
            sql = ""
            if new_field.null:
                sql = self.sql_alter_column_null % {
                    'column': self.quote_name(new_field.column)
                }
            else:
                sql = self.sql_alter_column_not_null % {
                    'column': self.quote_name(new_field.column)
                }
            self.execute(
                self.sql_alter_column % {
                    'table': self.quote_name(model._meta.db_table),
                    'changes': sql
                }
            )

        # Need to add check constraint
        if alter_field_check_constraint and new_db_field['check']:
            self.execute(
                self.sql_create_check % {
                    'table': self.quote_name(model._meta.db_table),
                    'name': self._create_index_name(model, [new_field.column], suffix="_check"),
                    'column': self.quote_name(new_field.column),
                    'check': new_db_field['check'],
                }
            )
        # Need to change incoming foreign key field type
        incoming_relations = []
        if alter_incomming_fk_data_type:
            incoming_relations.extend(new_field.model._meta.get_all_related_objects())

        # Need to add new PK
        if alter_field_primary_key and new_field.primary_key:
            # Drop old PK if available
            try:
                self.execute(
                    self.sql_drop_pk % {
                        'table': self.quote_name(model._meta.db_table)
                    }
                )
            except:
                pass
            self.__model = model
            self.execute(
                self.sql_create_pk % {
                    'table': self.quote_name(model._meta.db_table),
                    'name': self._create_index_name(model, [new_field.column], suffix="_pk"),
                    'columns': self.quote_name(new_field.column)
                }
            )
            # Need to update all incoming relations
            incoming_relations.extend(new_field.model._meta.get_all_related_objects())
        # Need to add a unique constraint
        elif alter_field_unique and new_field.unique:
            self.execute(
                self.sql_create_unique % {
                    'table': self.quote_name(model._meta.db_table),
                    'name': self._create_index_name(model._meta.db_table, [new_field.column], suffix="_uniq"),
                    'columns': self.quote_name(new_field.column),
                }
            )
            # Need to add a index
        elif alter_field_index and new_field.db_index:
            self.execute(
                self.sql_create_index % {
                    'table': self.quote_name(model._meta.db_table),
                    'name': self._create_index_name(model, [new_field.column], suffix="_index"),
                    'columns': self.quote_name(new_field.column),
                    'extra': "",
                }
            )
        # Update incoming FK field
        for inc_rel in incoming_relations:
            fk_db_field = inc_rel.field.db_parameters(connection=self.connection)
            fk_db_field_type = fk_db_field['type']
            sql = self.sql_alter_column_type % {
                'column': self.quote_name(inc_rel.field.column),
                'type': fk_db_field_type,
            }
            self.execute(
                self.sql_alter_column % {
                    'table': self.quote_name(inc_rel.model._meta.db_table),
                    'changes': sql
                }
            )

        # need to reorg table if we changed the field type of fk field
        if len(incoming_relations) > 0:
            self._reorg_tables()

        # Rebuild/make FK constraint, if it have any
        if (djangoVersion[0:2] < (1, 9)):
            if new_field.rel:
                self.execute(
                    self.sql_create_fk % {
                        'table': self.quote_name(model._meta.db_table),
                        'name': self._create_index_name(model, [new_field.column], suffix="_fk"),
                        'column': self.quote_name(new_field.column),
                        'to_table': self.quote_name(new_field.rel._meta.db_table),
                        'to_column': self.quote_name(new_field.rel.get_related_field().column),
                    }
                )
        else:
            if new_field.remote_field:
                self.execute(
                    self.sql_create_fk % {
                        'table': self.quote_name(model._meta.db_table),
                        'name': self._create_index_name(model, [new_field.column], suffix="_fk"),
                        'column': self.quote_name(new_field.column),
                        'to_table': self.quote_name(new_field.remote_field.model._meta.db_table),
                        'to_column': self.quote_name(new_field.remote_field.get_related_field().column),
                    }
                )
        # Rebuild incoming FK constraints
        if rebuild_incomming_fk:
            for inc_rel in new_field.model._meta.get_all_related_objects():
                self.execute(
                    self.sql_create_fk % {
                        'table': self.quote_name(inc_rel.model._meta.db_table),
                        'name': self._create_index_name(inc_rel.model, [inc_rel.field.column], suffix="_fk"),
                        'column': self.quote_name(inc_rel.field.column),
                        'to_table': self.quote_name(model._meta.db_table),
                        'to_column': self.quote_name(new_field.column),
                    }
                )

    def alterFieldDataTypeByRemaking(self, model, old_field, new_field, strict):
        tmp_new_field = copy.deepcopy(new_field)
        tmp_new_field.column = truncate_name("%s%s" % (self.psudo_column_prefix, tmp_new_field.column),
                                             self.connection.ops.max_name_length())
        self.add_field(model, tmp_new_field)

        # Transfer data from old field to new tmp field
        self.execute("UPDATE %s set %s=%s" % (
            self.quote_name(model._meta.db_table),
            self.quote_name(tmp_new_field.column),
            self.quote_name(old_field.column)
        )
                     )
        self.remove_field(model, old_field)
        return tmp_new_field, new_field

    def add_field(self, model, field):
        self.__model = model
        notnull = not field.null
        field.null = True
        p_key = field.primary_key
        field.primary_key = False
        unique = field.unique
        field._unique = False

        super(DB2SchemaEditor, self).add_field(model, field)
        if field.remote_field is not None and hasattr(field.remote_field, 'through'):
            rel_condition = field.remote_field.through._meta.auto_created
        else:
            rel_condition = False

        if isinstance(field, ManyToManyField) and rel_condition:
            return
        else:
            self._reorg_tables()
        sql = None
        if notnull or unique or p_key:
            del_column = self.sql_delete_column % {
                'table': self.quote_name(model._meta.db_table), 'column': self.quote_name(field.column)
            }
            if notnull:
                field.null = False
                sql = self.sql_alter_column_not_null % {'column': self.quote_name(field.column)}
                sql = self.sql_alter_column % {'table': self.quote_name(model._meta.db_table), 'changes': sql}
                try:
                    self.execute(sql)
                    self._reorg_tables()
                except Error as e:
                    self.execute(del_column)
                    raise e
            if p_key:
                field.primary_key = True
                cur = self.connection.cursor()
                # remove other pk if available
                for other_pk in cur.connection.primary_keys(True, cur.connection.get_current_schema(),
                                                            model._meta.db_table):
                    self.execute(
                        self.sql_delete_pk % {
                            'table': self.quote_name(model._meta.db_table),
                            'name': other_pk['PK_NAME']
                        }
                    )
                sql = self.sql_create_pk % {
                    'table': self.quote_name(model._meta.db_table),
                    'name': self._create_index_name(model, [field.column], suffix="_pk"),
                    'columns': self.quote_name(field.column)
                }
                try:
                    self.execute(sql)
                    self._reorg_tables()
                except Error as e:
                    self.execute(del_column)
                    raise e
            elif unique:
                field._unique = True
                constraint_name = self._create_index_name(model, [field.column], suffix="_uniq")
                sql = self.sql_create_unique % {
                    'table': self.quote_name(model._meta.db_table), 'name': constraint_name,
                    'columns': self.quote_name(field.column)
                }
                try:
                    self.execute(sql)
                    self._reorg_tables()
                except Error as e:
                    self.execute(del_column)
                    raise e

    def alter_db_table(self, model, old_db_table, new_db_table):
        super(DB2SchemaEditor, self).alter_db_table(model, old_db_table, new_db_table)

    def _alter_many_to_many(self, model, old_field, new_field, strict):
        deferred_constraints = {
            'pk': {},
            'unique': {},
            'index': {},
            'check': {}
        }

        if ((old_field.remote_field is not None and hasattr(old_field.remote_field, 'through')) and
                (new_field.remote_field is not None and hasattr(new_field.remote_field, 'through'))):
            old_field_rel_through = old_field.remote_field.through
            rel_old_field = old_field.remote_field.through._meta.get_field(old_field.m2m_reverse_field_name())
            rel_new_field = new_field.remote_field.through._meta.get_field(new_field.m2m_reverse_field_name())
        else:
            rel_old_field = None
            rel_new_field = None

        if ((rel_old_field is not None) and (rel_new_field is not None)):
            with self.connection.cursor() as cur:
                constraints = self.connection.introspection.get_constraints(cur, old_field_rel_through._meta.db_table)
            for constr_name, constr_dict in constraints.items():
                if constr_dict['foreign_key'] is not None:
                    self.execute(self.sql_delete_fk % {
                        "table": self.quote_name(old_field_rel_through._meta.db_table),
                        "name": constr_name,
                    })
            self._defer_constraints_check(constraints, deferred_constraints, rel_old_field, rel_new_field,
                                          old_field_rel_through, defer_pk=True, defer_unique=True, defer_index=True)

            self._reorg_tables()
            super(DB2SchemaEditor, self)._alter_many_to_many(model, old_field, new_field, strict)
            self._restore_constraints_check(deferred_constraints, rel_old_field, rel_new_field, new_field.rel.through)

    def _reorg_tables(self):
        checkReorgSQL = "select tabschema, tabname from sysibmadm.admintabinfo where reorg_pending = 'Y'"
        res = []
        reorgSQLs = []
        with self.connection.cursor() as cursor:
            cursor.execute(checkReorgSQL)
            res = cursor.fetchall()
        if res:
            for sName, tName in res:
                reorgSQL = '''CALL SYSPROC.ADMIN_CMD('REORG TABLE "%(sName)s"."%(tName)s"')''' % {
                    'sName': sName, 'tName': tName
                }
                reorgSQLs.append(reorgSQL)
        for sql in reorgSQLs:
            self.execute(sql)

    def _defer_constraints_check(self, constraints, deferred_constraints, old_field, new_field, model, defer_pk=False,
                                 defer_unique=False, defer_index=False, defer_check=False):
        for constr_name, constr_dict in constraints.items():
            if defer_pk and constr_dict['primary_key'] is True:
                if old_field.column in constr_dict['columns']:
                    self.execute(self.sql_delete_pk % {
                        'table': model._meta.db_table,
                        'name': constr_name
                    })
                    deferred_constraints['pk'][constr_name] = constr_dict['columns']
                    continue
            if defer_unique and constr_dict['unique'] is True:
                if old_field.column in constr_dict['columns']:
                    try:
                        self.execute(self.sql_delete_unique % {
                            'table': model._meta.db_table,
                            'name': constr_name
                        })
                        deferred_constraints['unique'][constr_name] = constr_dict['columns']
                        continue
                    except:
                        continue
            if defer_index and constr_dict['index'] is True:
                if old_field.column in constr_dict['columns']:
                    try:
                        self.execute(self.sql_delete_index % {
                            'table': model._meta.db_table,
                            'name': constr_name
                        })
                        deferred_constraints['index'][constr_name] = constr_dict['columns']
                    except:
                        pass
            if defer_check and constr_dict['check'] is True:
                if old_field.column in constr_dict['columns']:
                    self.execute(self.sql_delete_check % {
                        'table': model._meta.db_table,
                        'name': constr_name
                    })
                    deferred_constraints['check'][constr_name] = constr_dict['columns']

        return deferred_constraints

    def _restore_constraints_check(self, deferred_constraints, old_field, new_field, model):
        self.__model = model
        for pk_name, columns in deferred_constraints['pk'].items():
            self.execute(self.sql_create_pk % {
                'table': model._meta.db_table,
                'name': pk_name,
                'columns': ', '.join(column.replace(old_field.column, new_field.column) for column in columns)
            })
        for constr_name, columns in deferred_constraints['unique'].items():
            self.execute(self.sql_create_unique % {
                'table': model._meta.db_table,
                'name': constr_name,
                'columns': ', '.join(column.replace(old_field.column, new_field.column) for column in columns)
            })
        for index_name, columns in deferred_constraints['index'].items():
            self.execute(self.sql_create_index % {
                'table': model._meta.db_table,
                'name': index_name,
                'columns': ', '.join(column.replace(old_field.column, new_field.column) for column in columns),
                'extra': ""
            })

    def quote_value(self, value):
        if isinstance(value, (datetime.datetime, datetime.date, datetime.time, str)):
            escape_quotes = lambda s: f'{s}'.replace("'", "''")
            return f"'{escape_quotes(value)}'"
        elif isinstance(value, bool):
            return '1' if value else '0'
        elif isinstance(value, uuid.UUID):
            return f"'{value}'"
        elif isinstance(value, bytes):
            return f"BLOB(X'{value.hex()}')"
        elif isinstance(value, datetime.timedelta):
            # time intervals will be stored as double number of seconds
            return f"{value / datetime.timedelta(seconds=1)}"
        return str(value)



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                      |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi              |
# +--------------------------------------------------------------------------+

import sys

from django.db.models.sql import compiler

from itertools import zip_longest


class SQLCompiler(compiler.SQLCompiler):
    pass


class SQLInsertCompiler(compiler.SQLInsertCompiler, SQLCompiler):
    def as_sql(self):
        """Support returning identity val with single query."""
        (sql, params), *_ = super().as_sql()
        qn = self.connection.ops.quote_name
        opts = self.query.get_meta()
        sql = f'SELECT {qn(opts.pk.column)} FROM FINAL TABLE ({sql})'
        return [(sql, params)]


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    pass


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    pass


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    pass



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2014.                                      |
# +--------------------------------------------------------------------------+
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi, Steven James|
# +--------------------------------------------------------------------------+

import sys

from setuptools import setup, find_packages, find_namespace_packages
from distutils.core import setup, Extension

PACKAGE = 'django-pyodbc-iseries'
# VERSION = __import__('django_iseries').__version__
VERSION = '1.0.0'
LICENSE = 'Apache License 2.0'
# DESCRIPTION = __import__('django_iseries').__doc__
DESCRIPTION = ''

print(find_packages(where='src'))

setup(
    name=PACKAGE,
    version=VERSION,
    license=LICENSE,
    platforms='All',
    install_requires=['pyodbc>=4.0.27', 'django>=2.2.0', 'pytz', 'sqlparse'],
    dependency_links=['https://pypi.org/project/pyodbc/', 'http://pypi.python.org/pypi/Django/',
                      'https://www.ibm.com/support/pages/ibm-i-access-client-solutions'],
    description=DESCRIPTION,
    long_description=DESCRIPTION,
    author='Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi, Steven James',
    author_email='steven@waitforitjames.com',
    maintainer='Steven James',
    maintainer_email='steven@waitforitjames.com',
    url='https://github.com/soundstripe/django-pyodbc-iseries',
    keywords='django iseries backends adapter IBM Data Servers database db2',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    classifiers=['Development Status :: 4 - Beta',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: Apache Software License',
                 'Operating System :: Microsoft :: Windows :: Windows NT/2000',
                 'Operating System :: Unix',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: MacOS',
                 'Topic :: Database :: Front-Ends'],
    data_files=[('', ['./README.md']),
                ('', ['./CHANGES']),
                ('', ['./LICENSE'])],
    zip_safe=False,
    include_package_data=True,
)



# +--------------------------------------------------------------------------+
# |  Licensed Materials - Property of IBM                                    |
# |                                                                          |
# | (C) Copyright IBM Corporation 2009-2018.                                      |
# +--------------------------------------------------------------------------+
# | This module complies with Django 1.0 and is                              |
# | Licensed under the Apache License, Version 2.0 (the "License");          |
# | you may not use this file except in compliance with the License.         |
# | You may obtain a copy of the License at                                  |
# | http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable |
# | law or agreed to in writing, software distributed under the License is   |
# | distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY |
# | KIND, either express or implied. See the License for the specific        |
# | language governing permissions and limitations under the License.        |
# +--------------------------------------------------------------------------+
# | Authors: Ambrish Bhargava, Tarun Pasrija, Rahul Priyadarshi              |
# +--------------------------------------------------------------------------+

"""
DB2 database backend for Django.
Requires: ibm_db_dbi (http://pypi.python.org/pypi/ibm_db) for python
"""

from django.core.exceptions import ImproperlyConfigured
from django.db import utils
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.features import BaseDatabaseFeatures
from django.db.backends.base.validation import BaseDatabaseValidation

# Importing internal classes from iseries package.
from django_iseries.pybase import DatabaseWrapper as PyBaseDatabaseWrapper
from django_iseries.client import DatabaseClient
from django_iseries.creation import DatabaseCreation
from django_iseries.introspection import DatabaseIntrospection
from django_iseries.operations import DatabaseOperations
from django_iseries.schemaEditor import DB2SchemaEditor
from . import Database

dbms_name = 'dbname'


class DatabaseFeatures(BaseDatabaseFeatures):
    allows_group_by_pk = False
    allows_group_by_selected_pks = False

    interprets_empty_strings_as_nulls = False

    can_use_chunked_reads = True
    can_return_id_from_insert = True
    can_return_ids_from_bulk_insert = True
    uses_savepoints = True
    can_release_savepoints = True

    related_fields_match_type = False
    has_select_for_update = True
    has_select_for_update_skip_locked = True
    has_select_for_update_of = True
    select_for_update_of_column = True

    test_db_allows_multiple_connections = False

    supports_unspecified_pk = True

    supports_forward_references = False  # ??

    # Custom query class has been implemented
    # django.db.backends.db2.query.query_class.DB2QueryClass
    uses_custom_query_class = True

    # transaction is supported by DB2
    supports_transactions = True

    uppercases_column_names = True
    allows_auto_pk_0 = True
    can_defer_constraint_checks = False
    requires_rollback_on_dirty_transaction = True
    supports_regex_backreferencing = True
    supports_timezones = False
    has_bulk_insert = True

    supports_long_model_names = False
    can_distinct_on_fields = False
    supports_paramstyle_pyformat = False
    supports_sequence_reset = True
    # DB2 doesn't take default values as parameter
    requires_literal_defaults = True
    can_introspect_big_integer_field = True
    can_introspect_boolean_field = False
    can_introspect_positive_integer_field = False
    can_introspect_small_integer_field = True
    can_introspect_null = True
    can_introspect_ip_address_field = False
    can_introspect_time_field = True
    can_rollback_ddl = True
    has_case_insensitive_like = False
    bare_select_suffix = ' FROM SYSIBM.SYSDUMMY1'
    implied_column_null = True
    supports_select_for_update_with_limit = False
    ignores_table_name_case = True
    supports_over_clause = True

    create_test_procedure_without_params_sql = None
    create_test_procedure_with_int_param_sql = None


class DatabaseValidation(BaseDatabaseValidation):
    # Need to do validation for DB2 and ibm_db version
    def validate_field(self, errors, opts, f):
        pass


class DatabaseWrapper(BaseDatabaseWrapper):
    """
    This is the base class for DB2 backend support for Django.
    """
    data_types = {}
    vendor = 'db2iseries'
    display_name = 'Db2 for i (pyodbc)'
    operators = {
        "exact": "= %s",
        "iexact": "LIKE UPPER(%s) ESCAPE '\\'",
        "contains": "LIKE %s ESCAPE '\\'",
        "icontains": "LIKE UPPER(%s) ESCAPE '\\'",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
        "startswith": "LIKE %s ESCAPE '\\'",
        "endswith": "LIKE %s ESCAPE '\\'",
        "istartswith": "LIKE UPPER(%s) ESCAPE '\\'",
        "iendswith": "LIKE UPPER(%s) ESCAPE '\\'",
    }
    Database = Database

    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    validation_class = DatabaseValidation
    ops_class = DatabaseOperations

    # The patterns below are used to generate SQL pattern lookup clauses when
    # the right-hand side of the lookup isn't a raw string (it might be an expression
    # or the result of a bilateral transformation).
    # In those cases, special characters for LIKE operators (e.g. \, %, _)
    # should be escaped on the database side.
    #
    # Note: we use str.format() here for readability as '%' is used as a wildcard for
    # the LIKE operator.
    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\', '\\'), '%', '\%'), '_', '\_')"
    pattern_ops = {
        'contains': "LIKE '%' || {} || '%' ESCAPE '\\'",
        'icontains': "LIKE '%' || UPPER({}) || '%' ESCAPE '\\'",
        'startswith': "LIKE {} || '%' ESCAPE '\\'",
        'istartswith': "LIKE UPPER({}) || '%' ESCAPE '\\'",
        'endswith': "LIKE '%' || {} ESCAPE '\\'",
        'iendswith': "LIKE '%' || UPPER({}) ESCAPE '\\'",
    }

    error_codes_remap_to_database_error = ['28000']

    # Constructor of DB2 backend support. Initializing all other classes.
    def __init__(self, *args):
        super(DatabaseWrapper, self).__init__(*args)
        self.ops = DatabaseOperations(self)
        self.client = DatabaseClient(self)
        self.features = DatabaseFeatures(self)
        self.creation = DatabaseCreation(self)

        self.data_types = self.creation.data_types
        self.data_type_check_constraints = self.creation.data_type_check_constraints

        self.introspection = DatabaseIntrospection(self)
        self.validation = DatabaseValidation(self)
        self.databaseWrapper = PyBaseDatabaseWrapper()

    # Method to check if connection is live or not.
    def __is_connection(self):
        return self.connection is not None

    # To get dict of connection parameters 
    def get_connection_params(self):
        kwargs = {}

        settings_dict = self.settings_dict
        database_name = settings_dict['NAME']
        database_user = settings_dict['USER']
        database_pass = settings_dict['PASSWORD']
        database_host = settings_dict['HOST']
        database_port = settings_dict['PORT']
        database_options = settings_dict['OPTIONS']

        if database_name != '' and isinstance(database_name, str):
            kwargs['database'] = database_name
        else:
            raise ImproperlyConfigured("Please specify the valid database Name to connect to")

        if isinstance(database_user, str):
            kwargs['user'] = database_user

        if isinstance(database_pass, str):
            kwargs['password'] = database_pass

        if isinstance(database_host, str):
            kwargs['host'] = database_host

        if isinstance(database_port, str):
            kwargs['port'] = database_port

        if isinstance(database_host, str):
            kwargs['host'] = database_host

        if isinstance(database_options, dict):
            kwargs['options'] = database_options

        if (settings_dict.keys()).__contains__('PCONNECT'):
            kwargs['PCONNECT'] = settings_dict['PCONNECT']

        if 'CURRENTSCHEMA' in settings_dict:
            database_schema = settings_dict['CURRENTSCHEMA']
            if isinstance(database_schema, str):
                kwargs['currentschema'] = database_schema

        if 'SECURITY' in settings_dict:
            database_security = settings_dict['SECURITY']
            if isinstance(database_security, str):
                kwargs['security'] = database_security

        if 'SSLCLIENTKEYDB' in settings_dict:
            database_sslclientkeydb = settings_dict['SSLCLIENTKEYDB']
            if isinstance(database_sslclientkeydb, str):
                kwargs['sslclientkeydb'] = database_sslclientkeydb

        if 'SSLCLIENTKEYSTOREDBPASSWORD' in settings_dict:
            database_sslclientkeystoredbpassword: object = settings_dict['SSLCLIENTKEYSTOREDBPASSWORD']
            if isinstance(database_sslclientkeystoredbpassword, str):
                kwargs['sslclientkeystoredbpassword'] = database_sslclientkeystoredbpassword

        if 'SSLCLIENTKEYSTASH' in settings_dict:
            database_sslclientkeystash = settings_dict['SSLCLIENTKEYSTASH']
            if isinstance(database_sslclientkeystash, str):
                kwargs['sslclientkeystash'] = database_sslclientkeystash

        if 'SSLSERVERCERTIFICATE' in settings_dict:
            database_sslservercertificate = settings_dict['SSLSERVERCERTIFICATE']
            if isinstance(database_sslservercertificate, str):
                kwargs['sslservercertificate'] = database_sslservercertificate

        return kwargs

    # To get new connection from Database
    def get_new_connection(self, conn_params):
        connection = self.databaseWrapper.get_new_connection(conn_params)
        return connection

    # Over-riding _cursor method to return DB2 cursor.

    def create_cursor(self, name=None):
        return self.databaseWrapper._cursor(self.connection)

    def init_connection_state(self):
        pass

    def is_usable(self):
        if self.databaseWrapper.is_active(self.connection):
            return True
        else:
            return False

    def _set_autocommit(self, autocommit):
        self.connection.autocommit = autocommit

    def close(self):
        self.validate_thread_sharing()
        if self.connection is not None:
            self.databaseWrapper.close(self.connection)
            self.connection = None

    def get_server_version(self):
        if not self.connection:
            self.cursor()
        return self.databaseWrapper.get_server_version(self.connection)

    def schema_editor(self, *args, **kwargs):
        return DB2SchemaEditor(self, *args, **kwargs)

    def disable_constraint_checking(self):
        raise utils.NotSupportedError(
            "Db2 for iSeries currently supports no method of disabling constraints on a per-session basis."
        )

    def connect(self):
        try:
            super().connect()
        except Database.InterfaceError as e:
            """Django expects errors such as invalid password to be DatabaseError"""
            if e.args[0] in self.error_codes_remap_to_database_error:
                raise utils.DatabaseError(*e.args)
            raise
